#include "ni488.h"
#include <time.h>
#include <string.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "s.h"
#include <getopt.h>

/* Flags  */
static int portno = 12345;
static int help_flag = 0;

int parse(argc,argv);
int help();
int Device = 0;			/* Device unit descriptor                  */
int BoardIndex = 0;		/* Interface Index (GPIB0=0,GPIB1=1,etc.)  */
#define bufsize 2048

int main (argc, argv)
     int argc;
     char **argv;
{
  delay = 200;
  if (!parse(argc,argv))
    return -1;
  if (help_flag)
    return 0;
  
  if (verbose_flag)
    printf ("verbose flag is set\n");
  if (temp_flag && (verbose_flag==1 || debug==1))
    printf ("Temperature mode is set, No multiplexer operation possible!\n");
  if (verbose_flag==1 || debug==1)
    printf("Starting server...\n");

  setbuf(stdout,NULL);
  double rc, rorg;
  struct timespec tdel, trem;
  time_t curtime;
  struct tm *loctime;
  const char delimiters[] = " :\t";
  const char delimnul[] = "";
  char readstring[bufsize];
  char *token;
  char *cp;
  char c;
  static char scannerstring[stringl] = "";
  char timebuffer[stringl];
  int er, sockfd, newsockfd, clilen, icount;
  char buffer[bufsize];
  short chid, iaddr, ichannel, i;
  struct sockaddr_in serv_addr, cli_addr;
  int n; // , gpibid[32];
  int PrimaryAddress;           /* Primary address of the device           */
  int SecondaryAddress = 0;	/* Secondary address of the device         */
  for (i = 1; i <= 30; i++)
    {
      PrimaryAddress = i;
      gpibid[i] = ibdev (	/* Create a unit descriptor handle         */
			  BoardIndex,	/* Board Index (GPIB0 = 0, GPIB1 = 1, ...) */
			  PrimaryAddress,	/* Device primary address                  */
			  SecondaryAddress,	/* Device secondary address                */
			  T10s,	/* Timeout setting (T10s = 10 seconds)     */
			  1,	/* Assert EOI line at end of write         */
			  0);	/* EOS termination mode                    */
      if (ibsta & ERR)
	{			/*     Check for GPIB Error                    */
	  GpibError ("ibdev Error");
	}
      ibclr (gpibid[i]);	/* Clear the device                        */
      if (ibsta & ERR)
	{
	  GpibError ("ibclr Error");
	}
    }
  
  initchannels(); 
  initgpib(); 
  
  if (verbose_flag)
    printf("Opening socket on port %i\n",portno);
  sockfd = socket (AF_INET, SOCK_STREAM, 0);
  if (sockfd < 0)
    error ("ERROR opening socket");
  bzero ((char *) &serv_addr, sizeof (serv_addr));
  
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = INADDR_ANY;
  serv_addr.sin_port = htons (portno);
  if (bind (sockfd, (struct sockaddr *) &serv_addr, sizeof (serv_addr)) < 0)
    error ("ERROR on binding");
  listen (sockfd, 25);
  clilen = sizeof (cli_addr);

  do
    {
      newsockfd = accept (sockfd, (struct sockaddr *) &cli_addr, &clilen);
      if (verbose_flag==1 || debug==1)
	{
	  curtime = time(0);
	  printf(ctime(&curtime));
	  printf ("count %d fd %d \n", icount++, newsockfd);
	}
      if (newsockfd < 0)
	error ("ERROR on accept");
      bzero (buffer, bufsize);
      n = read (newsockfd, buffer, bufsize - 1);
      if (debug==1 || verbose_flag==1)
	{
	  printf ("%s length: %d \n", buffer, strlen (buffer));
	  printf ("Command: %s",buffer);
	}
      if (n < 0)
	error ("ERROR reading from socket");

      cp = strdup (buffer);	/* Make writable copy.  */
      chomp(cp);   /* Only work with part up to first newline (if any) */
      token = strtok (cp, delimiters);
      c = toupper (token[0]);

      token = strtok (NULL, delimiters);
      if (token)
	{
	  iaddr = atoi (token);
	  if (debug==1 || verbose_flag==1)
	    printf ("%c %d %s  \n", c, iaddr, token);
	}
      sprintf(readstring,  "\n"); 
      if (iaddr>0 || c == 'I' || c == 'Q'  || c == 'D' || c == 'T') 
	{ 
	  switch (c)
	    {
	      
	    case 'D':
	      if (debug)
		{
		  set_debug(0);
		  sprintf(readstring,  "Debug off\n"); 
		}
	      else
		{
		  set_debug(1);
		  sprintf(readstring,  "Debug on\n"); 
		}

	      n = write (newsockfd, readstring, strlen (readstring));
	      break;
	      
	    case 'T':
	      if (token != NULL)
		{
		  delay = atoi(token);
		}
	      sprintf(readstring,  "GPIB delay set to %i ms",delay); 
	      if (debug==1 || verbose_flag==1)
		printf("%s\n",readstring);
	      n = write (newsockfd, readstring, strlen (readstring));
	      break;

	    case 'I':
	      sprintf(readstring,  "Updating channel configurations...\n"); 
	      n = write (newsockfd, readstring, strlen (readstring));
	      initchannels(debug); 
	      break;
	      
	      
	    case 'C':
	      token = strtok (NULL, delimnul);
	      if (token != NULL)
		{
		  ibwrt (gpibid[iaddr], token, strlen (token));
		  tdel.tv_sec = 0;
		  tdel.tv_nsec = delay * 1000000;
		  er = nanosleep (&tdel, &trem);
		  memset (readstring, 0, bufsize);
		  ibrd (gpibid[iaddr], readstring, bufsize);
		}
	      else
		sprintf(readstring,  "Warning: No data to write, aborting!\n"); 
	      n = write (newsockfd, readstring, strlen (readstring));
	      break;
	      
	    case 'R':
	      
	      memset (readstring, 0, bufsize);
	      ibrd (gpibid[iaddr], readstring, bufsize);
	      n = write (newsockfd, readstring, strlen (readstring));
	      break;
	      
	    case 'W':
	      token = strtok (NULL, delimnul);
	      if (token != NULL)
		{
		  ibwrt (gpibid[iaddr], token, strlen (token));
		  n = write (newsockfd, token, strlen (token));
		  tdel.tv_sec = 0;
		  tdel.tv_nsec = delay * 1000000;
		  er = nanosleep (&tdel, &trem);
		}
	      else
		{
		  sprintf(readstring,  "Warning: No data to write, aborting!\n"); 
		  n = write (newsockfd, readstring, strlen (readstring));
		}
	      break;

	    case 'B':
	      token = strtok (NULL, delimiters);
	      chid = 0;
	      if (token != NULL)
		{
		  chid = atoi(token);
		}
	      token = strtok (NULL, delimiters);
	      int np = 20;
	      if (token != NULL)
		{
		  np = atoi(token);
		}
	      rc=-32768;
	      rorg=-32768;
	      if ((iaddr>0)&&(iaddr<31)&&(chid>100)&&(chid<nchannels) && (check_card_ch(iaddr,chid) != 0)) 
		{
		  burst2700_init(iaddr,chid,np);
		  curtime = time (NULL);
		  loctime = localtime (&curtime);
		  strftime (timebuffer, stringl, "%Y:%m:%d:%H:%M:%S", loctime);
		  sprintf(readstring,"%s %d  %d:%d Starting burst of %d points\n",timebuffer,curtime,iaddr,chid,np);
		  n = write (newsockfd, readstring, strlen (readstring),0);
		  int sl = (np * 15 / 2000) + 1;
		  sleep(sl);
		  int resnp = 0;
		  resnp = burst2700_result(iaddr,chid);
		  int x;
		  int org_nsec = tdel.tv_nsec;
		  tdel.tv_sec = 0;
		  tdel.tv_nsec = 2 * 1000000; /* 2 millisec */
		  for (x = 0; x != resnp; ++x)
		    {
		      sprintf(readstring,"%f %d:%d %s %f %f\n",burst[x].time,iaddr,chid,channeldata[iaddr][chid].measuretype,burst[x].raw,burst[x].value);
		      if (debug==1 || verbose_flag==1)
			{
			  printf("Sending data for point %d: %s\n",x,readstring);
			}
		      n = write (newsockfd, readstring, strlen (readstring),0);
		      if (n < 0)
			{
			  printf("Socket write failed!\n");
			  break;
			}
		      er = nanosleep (&tdel, &trem);
		    }
		  tdel.tv_nsec = org_nsec;
		  sprintf(readstring,"Sent %d data\n",x);
		}
	      else
		{
		  curtime = time (NULL);
		  loctime = localtime (&curtime);
		  strftime (timebuffer, stringl, "%Y:%m:%d:%H:%M:%S", loctime);
		  sprintf(readstring,  "%s %d  %d:%d CHANNEL_ERROR %f %f\n", timebuffer, curtime, iaddr, chid, rorg, rc);
		}
	      n = write (newsockfd, readstring, strlen (readstring),0);
	      break;

	    case 'K':
	      token = strtok (NULL, delimiters);
	      /*	      token = strtok (NULL, delimnul);*/
	      chid = 0;
	      if (token != NULL)
		{
		  chid = atoi(token);
		}
	      rc=-32768;
	      rorg=-32768;
	      if ((iaddr>0)&&(iaddr<31)&&(chid>100)&&(chid<nchannels) && (check_card_ch(iaddr,chid) != 0)) 
		{
		  rc = readchannel2700(iaddr, chid, &rorg);
		  curtime = time (NULL);
		  loctime = localtime (&curtime);
		  strftime (timebuffer, stringl, "%Y:%m:%d:%H:%M:%S", loctime);
		  sprintf(readstring,  "%s %d  %d:%d %s %f %f\n", timebuffer, curtime, iaddr, chid, channeldata[iaddr][chid].measuretype, rorg, rc);
		  n = write (newsockfd, readstring, strlen (readstring),0);
		}
	      else
		{
		  curtime = time (NULL);
		  loctime = localtime (&curtime);
		  strftime (timebuffer, stringl, "%Y:%m:%d:%H:%M:%S", loctime);
		  sprintf(readstring,  "%s %d  %d:%d CHANNEL_ERROR %f %f\n", timebuffer, curtime, iaddr, chid, rorg, rc);
		  n = write (newsockfd, readstring, strlen (readstring),0);
		}
	      break;


	    case 'V':
              token = strtok (NULL, delimiters);
              chid = 0;
              if (token != NULL)
                {
                  chid = atoi(token);
                }
              rc=-32768;
              rorg=-32768;
              if ((iaddr>0)&&(iaddr<31)&&(chid>100)&&(chid<nchannels) && (check_card_ch(iaddr,chid) != 0)) 
                {
                  char newconf[stringl];
                  strcpy(newconf,"volt:dc");
                  token = strtok (NULL, delimnul);
                  if (token != NULL)
                    {
                      strcpy(newconf,token);
                    }
                  rc = readchannel2700_conf(iaddr, chid, &rorg,newconf);
                  curtime = time (NULL);
                  loctime = localtime (&curtime);
                  strftime (timebuffer, stringl, "%Y:%m:%d:%H:%M:%S", loctime);
                  sprintf(readstring,  "%s %d  %d:%d %s %f %f\n", timebuffer, curtime, iaddr, chid, newconf, rorg, rc);
                  n = write (newsockfd, readstring, strlen (readstring),0);
                }
              else
                {
                  curtime = time (NULL);
                  loctime = localtime (&curtime);
                  strftime (timebuffer, stringl, "%Y:%m:%d:%H:%M:%S", loctime);
                  sprintf(readstring,  "%s %d  %d:%d CHANNEL_ERROR %f %f\n", timebuffer, curtime, iaddr, chid, rorg, rc);
                  n = write (newsockfd, readstring, strlen (readstring),0);
                }
              break;

	    case 'O':
	      ibwrt (gpibid[iaddr], "route:open:all", 14L); 
	      sprintf(readstring,  "All realys open\n"); 
	      n = write (newsockfd, readstring, strlen (readstring));
	      break;
	      
	    case 'Q':
	      printf ("%s\n", "quitting");
	      sprintf(readstring,  "Server shutting down!.\n"); 
	      n = write (newsockfd, readstring, strlen (readstring));
	      break;
	    }
	  if (debug==1 || verbose_flag==1)
	    {
	      printf("Result: %s",readstring);
	    }
	}
      if (debug==1 || verbose_flag==1)
	{
	  printf ("shutting down newsocket %d\n", shutdown (newsockfd, 2));
	}
      close (newsockfd);
    }
  while (c != 'Q');
  printf ("shutting down %d\n", shutdown (sockfd, 2));
  return 0;
  /* * Cleanup - Done only once at the end of your application.*/
  ibonl (Device, 0);		/* Take the device offline                 */
  if (ibsta & ERR)
    {
      GpibError ("ibonl Error");
    }

}


int parse(argc,argv)
     int argc;
     char **argv;
{
  int c;
  while (1)
    {
      static struct option long_options[] =
        {
          /* These options set a flag.   */
          {"verbose", 0, &verbose_flag, 1},
          {"brief", 0, &verbose_flag, 0},
          {"temp", 0, &temp_flag, 1},
          {"notemp", 0, &temp_flag, 0},
          {"help", 0, &help_flag, 1},
          /* These options don't set a flag.                                                       
             We distinguish them by their indices.  */
	  {"port", 1, 0, 0},
          {"board", 1, 0, 0},
          {"debug", 1, 0, 0},
          {0, 0, 0, 0}
        };
      /* getopt_long stores the option index here.   */
      int option_index = 0;

      c = getopt_long (argc, argv, "phtvd:", long_options, &option_index);
      /* Detect the end of the options.   */
      if (c == -1)
        break;
      switch (c)
        {
        case 0:
          /* If this option set a flag, do nothing else now.   */
          if (long_options[option_index].flag != 0)
            break;
          if (!strcmp(long_options[option_index].name,"board"))
            {
              BoardIndex = atoi(optarg);
              break;
            }
          if (!strcmp(long_options[option_index].name,"port"))
            {
              portno = atoi(optarg);
	      break;
	    }
          if (!strcmp(long_options[option_index].name,"debug"))
            {
              set_debug(atoi(optarg));
              printf("Debuglevel: %i\n",debug);
              break;
            }
          printf ("option %s", long_options[option_index].name);
          if (optarg)
            printf (" with arg %s", optarg);
          printf ("\n");
          break;

        case 't':
          temp_flag = 1;
          break;

        case 'h':
          help_flag = 1;
          break;

        case 'v':
          verbose_flag = 1;
          break;

        case 'p':
          portno = atoi(optarg);

        case 'd':
          set_debug(atoi(optarg));
          printf("Debuglevel: %i\n",debug);
          break;
        case '?':
          /* getopt_long already printed an error message.   */
          return 0;

        default:
          printf("Unknown option!\n");
          return 0;
	}
    }
  if (help_flag)
    return help();
  return 1;
}

int help()
{
  printf("\n\n   DTU-energy-conversion GPIB-server\n\nUsage:\n    gpib_socket_server [options] &\n(\
remember the fork if the server is to run continuously)\n\n");
  printf("Possible options:\n");
  printf("    --verbose         Enables verbose mode\n");
  printf("    --brief           Disables verbose mode (default)\n");
  printf("    --board [boardnr] Sets GPIB board nr (default = 0)\n");
  printf("                      This can be used if more than one\n");
  printf("                      GPIB card is installed as each\n");
  printf("                      board will need it's own server.\n");
  printf("                      Notice that each server must bind\n");
  printf("                      to a separate port (see --port option).\n");
  printf("    --port [portnr]   Sets port nr (default = 12345)\n");
  printf("    --debug [level}   Sets debug level (default is 0)\n");
  printf("                      Debug can  be enabled and disabled\n");
  printf("                      through the socket intrface\n");
  printf("    --temp            Enables internal temperature compensatio mode.\n");
  printf("                      Note that a server running in temperature\n");
  printf("                      mode can NOT be used as a multiplexer!\n");
  printf("                      Default is normal operation.\n");
  printf("    --notemp          Sets normal operation (default).\n");
  printf("    --help            Prints this help and exits.\n");
  printf("    -v                Same as --verbose.\n");
  printf("    -p                Same as --port [portnr].\n");
  printf("    -t                Same as --temp.\n");
  printf("    -h                Same as --help.\n");
  printf("    -d [level]        Same as --debug [level].\n");
  printf("\n\n");
  printf("Note that internal temperature compensation on the Keithley2700\nis incompatible with using the 2700 as a multiplexer\n");
  printf("apart form the normal channel scanneing mode (that is it can NOT\nbe used as an impedance multiplexer in temperature mode!)\n\n");
  return 1;
}



